#pragma once

#include "componentbasic.h"
#include <messagebasic.h>
#include <QJsonObject>
#include <QJSEngine>

class MessageLoader;
class CompareKernel;
class ConvertKernel;
class LogicalNode;

/// <summary>
/// ͨ
/// </summary>
struct COMPONENTBASIC_EXPORT GeneralData : public TopicData {
	QJsonObject _element_bind;

	QString topicString() const override;
	std::shared_ptr<Serializable> newDefault() const override;

	void recoveryFrom(const QJsonObject& obj) override;
	void saveTo(QJsonObject& obj) const override;
};
// ==================================================================================

/// <summary>
/// Ϊʵڵ
/// </summary>
class COMPONENTBASIC_EXPORT MapKernal : virtual public Serializable {
private:
	/// <summary>
	/// 
	/// </summary>
	std::shared_ptr<MessageLoader> _bind_loader;

	/// <summary>
	/// {Topicdefault-value}
	/// </summary>
	QHash<QString, std::shared_ptr<TopicData>> _variable_types;

	/// <summary>
	/// Ƚ
	/// </summary>
	QHash<QString, std::shared_ptr<CompareKernel>> _compare_types;
	/// <summary>
	/// ת
	/// </summary>
	QHash<QString, std::shared_ptr<ConvertKernel>> _convert_types;

	/// <summary>
	/// пõĽڵ
	/// </summary>
	QHash<QString, std::shared_ptr<LogicalNode>> _logicalnode_types;

public:
	MapKernal(std::shared_ptr<MessageLoader> ins);
	virtual ~MapKernal() = default;

	/// <summary>
	/// ȡMessageLoader
	/// </summary>
	/// <returns></returns>
	std::shared_ptr<MessageLoader> messageLoader() const;

	// ͹ ======================================================
	/// <summary>
	/// ע
	/// </summary>
	/// <param name="init"></param>
	virtual void registData(std::shared_ptr<TopicData> init);
	/// <summary>
	/// ȡп
	/// </summary>
	/// <returns></returns>
	virtual QList<QString> dataTypes() const;
	/// <summary>
	/// ȡָtopic
	/// </summary>
	/// <param name="topic"></param>
	/// <returns></returns>
	virtual std::shared_ptr<TopicData> getData(const QString &topic);



	// Ƚ͹ ======================================================
	/// <summary>
	/// עȽ
	/// </summary>
	/// <param name="ins"></param>
	virtual void registCompare(std::shared_ptr<CompareKernel> ins);
	/// <summary>
	/// ȡȽ
	/// </summary>
	/// <returns></returns>
	virtual QList<QString> compareTypes() const;
	/// <summary>
	/// ȡȽʵ
	/// </summary>
	/// <param name="name"></param>
	/// <returns></returns>
	virtual std::shared_ptr<CompareKernel> getCompare(const QString& name);



	// ת͹ ======================================================
	/// <summary>
	/// עת
	/// </summary>
	/// <param name="ins"></param>
	virtual void registConvert(std::shared_ptr<ConvertKernel> ins);
	/// <summary>
	/// ת
	/// </summary>
	/// <returns></returns>
	virtual QList<QString> convertTypes() const;
	/// <summary>
	/// ȡָ͵ת
	/// </summary>
	/// <param name="name"></param>
	/// <returns></returns>
	virtual std::shared_ptr<ConvertKernel> getConvert(const QString& name);


	// ͨ Serializable ̳
	std::shared_ptr<Serializable> newDefault() const override;
	void recoveryFrom(const QJsonObject& obj) override{}
	void saveTo(QJsonObject& obj) const override{}
};

/// <summary>
/// ȽϹϵʵ
/// </summary>
class COMPONENTBASIC_EXPORT CompareKernel {
private:
	QJSEngine& _script_engine;
	QString _function_name;

public:
	CompareKernel(QJSEngine& bind_engine, const QString& func);

	QString name() const;
	bool compare(std::shared_ptr<TopicData> vleft, std::shared_ptr<TopicData> vright);
};

/// <summary>
/// תʵ
/// </summary>
class COMPONENTBASIC_EXPORT ConvertKernel {
public:
	typedef QString DATA_TOPIC_STRING;

	/// <summary>
	/// ת
	/// </summary>
	/// <param name="host">͹</param>
	/// <param name="bind_engine"></param>
	/// <param name="func"></param>
	ConvertKernel(std::shared_ptr<MapKernal> host, QJSEngine& bind_engine, const QString& func);
	/// <summary>
	/// תƣJS
	/// </summary>
	/// <returns></returns>
	QString name() const;

	/// <summary>
	/// Ͱ󶨱
	/// </summary>
	/// <returns>list{pair{,}}</returns>
	QList<std::pair<DATA_TOPIC_STRING, QString>> inputTable() const;
	/// <summary>
	/// ͱ
	/// </summary>
	/// <param name="table">list{pair{,}}</param>
	void inputReset(const QList<std::pair<DATA_TOPIC_STRING, QString>> &table);
	/// <summary>
	/// 
	/// </summary>
	/// <returns>pair{,}</returns>
	std::pair<DATA_TOPIC_STRING, QString> outputVariable() const;
	/// <summary>
	/// 
	/// </summary>
	/// <param name="appoint"></param>
	void outputReset(std::pair<DATA_TOPIC_STRING, QString> appoint);

	/// <summary>
	/// ִת
	/// </summary>
	/// <param name="input_variables">б</param>
	/// <returns></returns>
	std::shared_ptr<TopicData> convert(const QList<std::shared_ptr<TopicData>>& input_variables);

private:
	std::shared_ptr<MapKernal> _host_bind;

	QJSEngine& _script_engine;
	QString _function_name;

	QList<std::pair<DATA_TOPIC_STRING, QString>> _input_table;
	std::pair<DATA_TOPIC_STRING, QString> _output_appoint;
};




/// <summary>
/// ִн
/// </summary>
enum class LogicalResult {
	UNDEFINED,	// δ
	SUCCESS,	// ɹ
	FAILURE		// ʧ
};
/// <summary>
/// ߼ڵ
/// </summary>
enum class NodeKind {
	MAPNODE,		// ߼ڵ
	SEQUENCENODE,	// ˳ڵ
	SELECTORNODE,	// ѡڵ
	PARALLELNODE,	// нڵ
	COMPARENODE,	// ȽϽڵ
	ACTIONNODE,		// ڵ
};
/// <summary>
/// ߼ڵĻ
/// </summary>
class COMPONENTBASIC_EXPORT LogicalNode : virtual public Serializable {
private:
	int _node_id = 0;
	NodeKind _node_type = NodeKind::ACTIONNODE;
	std::weak_ptr<LogicalNode> _parent_bind;
	QList<std::shared_ptr<LogicalNode>> _child_list;

public:
	LogicalNode(NodeKind t = NodeKind::ACTIONNODE);
	virtual ~LogicalNode() = default;

	/// <summary>
	/// ȡڵ
	/// </summary>
	/// <returns></returns>
	NodeKind nodeKind() const;

	/// <summary>
	/// øڵ
	/// </summary>
	/// <param name="pnode"></param>
	void setParent(std::weak_ptr<LogicalNode> pnode);
	/// <summary>
	/// ˽ڵ
	/// </summary>
	/// <returns></returns>
	std::weak_ptr<LogicalNode> parent() const;
	/// <summary>
	/// ȡͼڵ
	/// </summary>
	/// <returns></returns>
	std::shared_ptr<LogicalNode> bindMap() const;

	/// <summary>
	/// ýڵid
	/// </summary>
	/// <param name="unique_id"></param>
	void setID(int unique_id);
	/// <summary>
	/// ȡڵid
	/// </summary>
	/// <returns></returns>
	int getID() const;
	/// <summary>
	/// ʵڵ
	/// </summary>
	/// <returns></returns>
	QString runtimeName() const;

	/// <summary>
	/// ڲ״̬
	/// </summary>
	virtual void reset() = 0;
	/// <summary>
	/// ȡڵ
	/// </summary>
	/// <returns></returns>
	virtual QString typeName() const = 0;
	/// <summary>
	/// ִ߼ڵ
	/// </summary>
	/// <param name="core"></param>
	/// <returns></returns>
	virtual LogicalResult execute() = 0;
	/// <summary>
	/// ڲʶݽӿ
	/// </summary>
	/// <returns>map{name, type-ins}</returns>
	virtual QHash<QString, std::shared_ptr<TopicData>> inputList() const = 0;
	/// <summary>
	/// ڲʶݽӿ
	/// </summary>
	/// <returns>map{name, type-ins}</returns>
	virtual QHash<QString, std::shared_ptr<TopicData>> outputList() const = 0;

	/// <summary>
	/// ½Ĭʵ
	/// </summary>
	/// <returns></returns>
	virtual std::shared_ptr<Serializable> newDefault() const = 0;
};

/// <summary>
/// Ͻڵ
/// </summary>
class COMPONENTBASIC_EXPORT ComposeNode {
private:
	QList<std::shared_ptr<LogicalNode>> _child_list;

public:
	/// <summary>
	/// ȡӽڵб
	/// </summary>
	/// <returns></returns>
	virtual QList<std::shared_ptr<LogicalNode>> children() const;
	/// <summary>
	/// ӽڵ
	/// </summary>
	/// <param name="node"></param>
	/// <param name="index"></param>
	virtual void insert(std::shared_ptr<LogicalNode> node, int index = -1);
	/// <summary>
	/// Ƴӽڵ
	/// </summary>
	/// <param name="node"></param>
	virtual void remove(std::shared_ptr<LogicalNode> node);

	/// <summary>
	/// ȡִڽڵ
	/// </summary>
	/// <returns></returns>
	virtual QList<std::shared_ptr<LogicalNode>> getForwards() const = 0;
	/// <summary>
	/// нڵִ״̬ط
	/// </summary>
	/// <param name="node"></param>
	/// <param name="ste"></param>
	/// <returns>ִб־</returns>
	virtual bool fallback(std::shared_ptr<LogicalNode> node, LogicalResult ste) = 0;
};

/// <summary>
/// ߼ڵ
/// </summary>
enum class IO_TYPE {
	INPUT, OUTPUT
};
/// <summary>
/// ԶΪڵڵʵ
/// </summary>
class COMPONENTBASIC_EXPORT BehaviorMapNode
	: public MapKernal, public LogicalNode, public ComposeNode {
private:
	/// <summary>
	/// ͼ
	/// </summary>
	QHash<QString, std::pair<IO_TYPE, std::shared_ptr<TopicData>>> _variables;
	QString _map_name;
	LogicalResult _state_value;

public:
	BehaviorMapNode(std::shared_ptr<MessageLoader> core);


	/// <summary>
	/// ñʹ׳UniException쳣
	/// ڵʼʱᰴעĬֵ
	/// </summary>
	virtual void setVariable(const QString& key, IO_TYPE t, std::shared_ptr<TopicData> ins);
	/// <summary>
	/// ȡָƵݱ
	/// </summary>
	/// <param name="key"></param>
	/// <returns>ָݣnullptrͲȷ</returns>
	virtual std::shared_ptr<TopicData> getVariable(const QString& key) const;


	/// <summary>
	/// Ϊڵ
	/// </summary>
	/// <param name="val"></param>
	virtual void resetName(const QString& val);
	/// <summary>
	/// ִ״̬
	/// </summary>
	void reset() override;
	/// <summary>
	/// ȡڵ
	/// </summary>
	/// <returns></returns>
	QString typeName() const override;

	/// <summary>
	/// ȡǰýڵ
	/// </summary>
	/// <returns></returns>
	QList<std::shared_ptr<LogicalNode>> getForwards() const override;
	/// <summary>
	/// ״̬
	/// </summary>
	/// <param name="node"></param>
	/// <param name="ste"></param>
	/// <returns></returns>
	bool fallback(std::shared_ptr<LogicalNode> node, LogicalResult ste) override;

	/// <summary>
	/// ִ״̬
	/// </summary>
	/// <returns></returns>
	LogicalResult execute() override;

	/// <summary>
	/// б
	/// </summary>
	/// <returns></returns>
	QHash<QString, std::shared_ptr<TopicData>> inputList() const override;
	/// <summary>
	/// б
	/// </summary>
	/// <returns></returns>
	QHash<QString, std::shared_ptr<TopicData>> outputList() const override;

	/// <summary>
	/// ʵ
	/// </summary>
	/// <param name="p"></param>
	/// <returns></returns>
	virtual std::shared_ptr<Serializable> newDefault() const override;
	void recoveryFrom(const QJsonObject& obj) override;
	void saveTo(QJsonObject& obj) const override;
};

// Ϊڵ ======================================================

/// <summary>
/// ˳ڵ
/// </summary>
class COMPONENTBASIC_EXPORT SequenceNode
	: public ComposeNode, public LogicalNode {
private:
	QHash<int, LogicalResult> _state_map;

public:
	SequenceNode();

	QList<std::shared_ptr<LogicalNode>> getForwards() const override;
	QString typeName() const override;

	LogicalResult execute() override;
	QHash<QString, std::shared_ptr<TopicData>> inputList() const override;
	QHash<QString, std::shared_ptr<TopicData>> outputList() const override;

	void reset() override;
	bool fallback(std::shared_ptr<LogicalNode> node, LogicalResult ste) override;

	/// <summary>
	/// ʵ
	/// </summary>
	/// <param name="p"></param>
	/// <returns></returns>
	virtual std::shared_ptr<Serializable> newDefault() const override;

	void recoveryFrom(const QJsonObject& obj) override;
	void saveTo(QJsonObject& obj) const override;
};

/// <summary>
/// ѡڵ
/// </summary>
class COMPONENTBASIC_EXPORT SelectorNode
	: public ComposeNode, public LogicalNode {
private:
	QHash<int, LogicalResult> _state_map;

public:
	SelectorNode();

	QList<std::shared_ptr<LogicalNode>> getForwards() const override;
	QString typeName() const override;

	LogicalResult execute() override;
	QHash<QString, std::shared_ptr<TopicData>> inputList() const override;
	QHash<QString, std::shared_ptr<TopicData>> outputList() const override;

	void reset() override;
	bool fallback(std::shared_ptr<LogicalNode> node, LogicalResult ste) override;

	/// <summary>
	/// ʵ
	/// </summary>
	/// <param name="p"></param>
	/// <returns></returns>
	virtual std::shared_ptr<Serializable> newDefault() const override;

	void recoveryFrom(const QJsonObject& obj) override;
	void saveTo(QJsonObject& obj) const override;
};

/// <summary>
/// нڵ
/// </summary>
class COMPONENTBASIC_EXPORT ParallelNode
	: public ComposeNode, public LogicalNode {
private:
	QHash<int, LogicalResult> _state_map;

public:
	ParallelNode();

	QHash<QString, std::shared_ptr<TopicData>> inputList() const override;
	QHash<QString, std::shared_ptr<TopicData>> outputList() const override;

	QString typeName() const override;

	QList<std::shared_ptr<LogicalNode>> getForwards() const override;
	bool fallback(std::shared_ptr<LogicalNode> node, LogicalResult ste) override;

	LogicalResult execute() override;
	void reset() override;

	/// <summary>
	/// ʵ
	/// </summary>
	/// <param name="p"></param>
	/// <returns></returns>
	virtual std::shared_ptr<Serializable> newDefault() const override;

	void recoveryFrom(const QJsonObject& obj) override;
	void saveTo(QJsonObject& obj) const override;
};

/// <summary>
/// ȽϽڵ
/// </summary>
class COMPONENTBASIC_EXPORT CompareNode : public LogicalNode {
private:
	std::shared_ptr<CompareKernel> _bind_delegate;
	QHash<QString, std::shared_ptr<TopicData>> _data_map;

public:
	CompareNode();

	/// <summary>
	/// ί
	/// </summary>
	/// <returns></returns>
	QString delegateName() const;
	/// <summary>
	/// 󶨱Ƚί
	/// </summary>
	/// <param name="ins"></param>
	void bindDelegate(std::shared_ptr<CompareKernel> ins);

	QHash<QString, std::shared_ptr<TopicData>> inputList() const override;
	QHash<QString, std::shared_ptr<TopicData>> outputList() const override;

	QString typeName() const override;
	LogicalResult execute() override;

	void reset() override;

	/// <summary>
	/// ʵ
	/// </summary>
	/// <param name="p"></param>
	/// <returns></returns>
	virtual std::shared_ptr<Serializable> newDefault() const override;

	void recoveryFrom(const QJsonObject& obj) override;
	void saveTo(QJsonObject& obj) const override;
};

// Ϊڵ ====================================================================

/// <summary>
/// Ϊ
/// </summary>
class COMPONENTBASIC_EXPORT BehaviorPerformPlugin :
public RespondRoute<
	WsRespond<EntityInitialRequest, RespondDefault>,
	WsRespond<EntityPreparedRequest, RespondDefault>,
	WsRespond<SyncRequest, RespondDefault>,
	WsRespond<DeduceRequest, RespondDefault>>,
public RequestRoute<> 
{
private:
	std::weak_ptr<WsEntity> _bind_entity;

public:
	void mapReset(const QString &path);

	// ͨ ProcList ̳
	void execute(std::shared_ptr<const EntityInitialRequest> in, QList<std::shared_ptr<RespondDefault>>& out) override;
	void execute(std::shared_ptr<const EntityPreparedRequest> in, QList<std::shared_ptr<RespondDefault>>& out) override;
	void execute(std::shared_ptr<const SyncRequest> in, QList<std::shared_ptr<RespondDefault>>& out) override;
	void execute(std::shared_ptr<const DeduceRequest> in, QList<std::shared_ptr<RespondDefault>>& out) override;

	void recoveryFrom(const QJsonObject& obj) override;
	void saveTo(QJsonObject& obj) const override;

	std::shared_ptr<Serializable> newDefault() const override;
	void bindEntity(std::weak_ptr<WsEntity> host) override;
	QString name() const override;
};

